package GGI::Color;

=head1 NAME

GGI::Color - Perl object class for arrays of LibGGI ggi_color structures

=head1 SYNOPSIS

  use GGI::Color;

  $color = new GGI::Color;
  $color = new GGI::Color $numcolors;

  $numcolors = $color->size;
  $color->size($numcolors);

  @shuffled_color_channels = $color->get();
  @red_color_channel = $color->get('r');
  @green_color_channel = $color->get('g');
  @blue_color_channel = $color->get('b');
  @alpha_color_channel = $color->get('a');
  @shuffled_color_channels_just_rgb = $color->get('rgb');

  @color_channels = $color->get($channel_select, $idx);
  @color_channels = $color->get($channel_select, $idx, $len);

  # Example: get red, green, blue channel values of second color.
  ($red, $green, $blue) = $color->get(rgb => 1, 1);

  $color->set($channel_select, \@color_channels);
  $color->set($channel_select, $idx, \@color_channels);
  $color->set($channel_select, $idx, $num, \@color_channels);

  # Example: set red, green, blue channel values of first color.
  $color->set(rgb => [$red, $green, $blue]);


=head1 DESCRIPTION

The GGI::Color object implements a simple interface to LibGGI's
ggi_color structure, or arrays thereof.

=cut

use strict;
use Carp;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);

require Exporter;
require DynaLoader;
require AutoLoader;

@ISA = qw(AutoLoader Exporter DynaLoader);

=head1 Exported constants

COLOR_PRECISION

=cut

@EXPORT = qw(COLOR_PRECISION);
$VERSION = '0.03';

bootstrap GGI::Color $VERSION;

=head1 Methods

GGI::Color objects support the following methods:

=head2 new

The GGI::Color object constructor, with no arguments, will allocate
space for a single LibGGI ggi_color structure.  Optionally a number 
of LibGGI ggi_color objects to allocate may be passed as a parameter;
they will be allocated consecutive in memory.

=cut

sub new {
    my $class = shift();
    my($this) = {};
    bless $this, $class;
    my($arg) = (@_ + 0) ? shift() : 1;
    $this->{obj} = $this->_calloc($arg);
    return undef unless ($this->{obj});
    $this->{num} = $arg;
    return $this;
}

sub DESTROY {
    my $this = shift();
    $this->_free($this->{obj});
}

=head2 size

The number of ggi_color structures stored in a GGI::Color object may 
be ascertained by calling the size method with no arguments.

When an argument is provided the GGI::Color's memory allocation
is reallocated to the given number of LibGGI ggi_color structures
and the new size is returned.

=cut

sub size {
    my $this = shift();
    if (@_ + 0) {
	my($num) = shift();
	unless ($this->{obj}) {
	    $this->{num} = $num;
	    $this->{obj} = $this->_calloc($num);
	    unless ($this->{obj}) {
		$this->{num} = 0 ;
		$this->{obj} = undef;
		return 0;
	    }
	}
	unless ($num == $this->{num}) {
	    $this->{obj} = $this->_realloc($this->{obj}, $num, $this->{num});
	    return $num if $this->{num} == $num;
	    $this->{num} = $num;
	}
	return $num;
    }
    return $this->{num};
}

=head2 get

Without an argument, returns an array consisting of integers representing
the values in each of the four color channels, four values for each 
ggi_color structure stored in the GGI::color object, in the order 
red, green, blue, alpha.

An optional first argument may select which color channels to include
in the array and how to order them.  This argument should be a string 
consisting only of the characters r, g, b or a, which represent the
red, green, blue and alpha channels respectively.  (y, u, and v will be
accepted in future implementations.)  The string may be any
length, and characters may be repeated.

If a second argument is supplied, it will be evaluated as the starting
index into the GGI::Color object; channel values will be returned 
for all ggi_color structures in the object including or after the given 
index.  The index of the first color is 0.  A negative index value
is interperated as an offset from the end of the GGI::Color's store
of ggi_color structures.

If a third argument is supplied, it will be evaluated as the number of 
ggi_color structures to return values for.  A negative value is 
interperated as a request to start at the index and move backwards,
thus the following two array values are the same:

   reverse($col->get('r'));
   $col->get('r', -1, 0 - $col->size);

The behavior if the requested number differs from the number of 
ggi_color structures actually stored in the GGI::color object is for 
the returned array to only contain values for the available ggi_color 
structures, and for the GGI::color object to retain the same size.  
This behavior may be modified by using the GGI::AutoColor subclass instead.

More arguments may be provided; this results in the equivalent of
calling the get method more than once and concatinating the results.
For example the following invocation would return the red channel
values of the ggi_color structures at indexes 3, 4, and 5, followed by 
the green channel values of those at indexes 3, 4, and 5, followed by 
the blue channel values of all starting at index 3.  Note the
use of the aternative list separator to make the channel select
string a bareword also improves readability.

    $color->get(r => 3, 3, g => 3, 3, b => 3);

By supplying an explicit undef for the starting index, processing
may be made to pick up where the last request left off (at the 
ggi_color structure after the one who's values were last read).  If 
an explicit undef is supplied as the channel select string, the
string last used is passed along.  An explicit undef in the 
length field causes the maximum number of ggi_color structures to 
be taken, through to the end of the GGI::color object.

=cut

sub get {
    my($this) = shift();
    return() if $this->size <= 0;
    my(@res);
    my($csel, $idx, $len) = ('rgba', 0, undef);
    while (@_ + 0) {
	my($arg) = shift();
	$csel = $arg if defined ($arg);
	next unless (@_ + 0);
	my($arg) = shift();
	if (defined($idx)) {
	    $idx = $arg; 
	    $idx += $this->size if ($idx < 0);
	}
	next unless (@_ + 0);
	$len = shift();
    } continue {
	if ($len < 0) {
	    if ($idx >= $this->size) {
		$len -= $this->size - $idx - 1;
		if ($len < 0) {
		    push(@res, 
			 $this->_get_rev($this->{obj}, $csel, 
					 ($this->size - 1), 
					 (($idx + $len < -1) ? 
					  ($idx + 1) : -$len)
					 )
			 );
		}
		$idx = $this->size - 1;
	    }
	    elsif ($idx >= 0) {
		push(@res, 
		     $this->_get_rev($this->{obj}, $csel, $idx, 
				     (($idx + $len < -1) ? ($idx + 1) : -$len)
				     )
		     );
	    }
	}
	else {
	    $len = $this->size unless defined($len);
	    if ($idx < 0) {
		$len += $idx;
		if ($len > 0) {
		    push(@res, 
			 $this->_get_fwd($this->{obj}, $csel, 0, 
					 (($len > $this->size) ? 
					  $this->size : $len)
					 )
			 );
		}
		$idx = 0;
	    }
	    elsif ($idx < $this->size) {
		push(@res, 
		     $this->_get_fwd($this->{obj}, $csel, $idx, 
				     (($idx + $len > $this->size) ? 
				      ($this->size - $idx) : $len)
				     )
		     );
	    }
	}
	$idx += $len;
    }
    return(@res);
}

=head2 set

In it's simplest form, the set method takes a reference to an array and 
places from the array the values of the red, green, blue and alpha channels,
respectively, into the first ggi_color structure in the GGI::Color object.  
If there are more than four values in the array, they are placed in the 
next ggi_color structure in the object, and so on.

The array reference may optionally be preceded by one two, or three 
parameters which may not themselves be array references.  

The first of these parameters acts as a channel selection mechanism, 
and should be a string consisting only of the characters 'r', 'g', 'b', 
and 'a'.  For example, the string 'rg' would put the values in the array 
that follows it into the red and green channels of the GGI::Color 
object's array of ggi_color structures, with alternating values going 
into the red and green channels until the array runs out of values
or the end of the GGI::Color object is reached.

The second parameter, if present, will select which ggi_color structure
in the GGI::Color object to start placing values in.  A negative value is
interperated as an offset from the end of the GGI::color object.

The third will limit the number of values which will be taken from the 
array.  Note this differs from the get method as it does not 
specify the number of ggi_color objects to alter.  That is, the following
statement would only alter the red and blue channels of the first two 
ggi_color structures in $col, leaving the array values 3,3,4,4 unused:

    $col->set(rb => 0, 4, [1, 1, 2, 2, 3, 3, 4, 4]);

A negative value will cause the operation to work backwards from the 
specified ggi_color structure.

Multiple array references may be supplied, and any of them may have
preceding parameters as described above.  The effect is the same as if the
set method was called multiple times. The channel selection value and the
number of colors to process may be reused (set to the value used
with the preceding array) by supplying an explicit undef.  If the starting 
index is not defined, processing continues at the next location. 

For example, the following invocation would cause all the color
channels in the first three ggi_color structures stored in $col to be 
set to the same values (1 in the red channel, 2 in the green channel,
3 in the red channel, and 4 in the alpha channel), but will 
not set the alpha channel of the fourth ggi_color structure to 5.

$col->set(r=>[1,1,1], bga=>1,[2,3,4,2,3,4], abg=>0,3,[4,2,3,5]); 

If both the starting index and the channel selection value are left 
undefined, the behavior is undefined for all but the first array
reference (see above).  This condition is reserved for possible 
future expansion.

It should be noted by those writing APIs or object classes which other 
people may use that versions of perl prior to 5.005 do not implement 
a way to determine the size of a tied array, so explicitly specifying 
the number of values to take would be wise, else your code will fail
when handed a tied array.  No attempt will be made to find the length
of an array if it has a defined length parameter associated with it.

=cut

sub set {
    my $this = shift();
    my ($csel, $idx, $len) = ('rgba', 0, undef);
    my($arg);
    while (@_ + 0) {
	$arg = shift();
	(($idx, $len) = (0, undef)), next
	    if (ref($arg) and UNIVERSAL::isa($arg, "ARRAY"));
	($csel) = ($arg) if defined($arg);
	(($idx, $len) = (0, undef)), next unless (@_ + 0);
	$arg = shift();
	(($idx, $len) = (0, undef)), next
	    if (ref($arg) and UNIVERSAL::isa($arg, "ARRAY"));
	if (defined($arg)) {
	    $idx = $arg;
	    $idx += $this->size if ($idx < 0);
	}
	($len = undef), next unless (@_ + 0);
	$arg = shift();
	($len = undef), next if (ref($arg) and UNIVERSAL::isa($arg, "ARRAY"));
	$len = $arg;
        croak "Expected array reference" unless(@_ +0);
	$arg = shift();
	next if (ref($arg) and UNIVERSAL::isa($arg, "ARRAY"));
	croak "Expected array reference";
    } continue {
	my($cnum) = length($csel);
	if($len < 0) {
	    $idx = 
		$this->_set_rev($this->{obj}, $csel, $idx, 
				($this->size - 1), $len, $arg);
	}
	else {
	    $len = defined($len) ? $len : (@$arg + 0);
	    $idx = 
		$this->_set_fwd($this->{obj}, $csel, $idx, 
				($this->size - 1), $len, $arg);
	}
    }
}

1;
__END__

=head1 AUTHOR

Brian S. Julin <bri@calyx.com>

=head1 SEE ALSO

GGI::(3pm), perl(1).

=cut
